﻿using System;
using System.Diagnostics.Contracts;
using System.Linq;
using System.Reflection;

namespace AgileDesign.Utilities
{
    public static class ConvertExtensions
    {
        private const BindingFlags BindingFlagsToUse = BindingFlags.NonPublic | BindingFlags.Public | BindingFlags.Instance;

        public static T ConvertTo<T>(this object value)
        {
            return (T)value.ConvertTo(typeof(T));
        }

        public static object ConvertTo
            (
            this object value,
            Type targetType
            )
        {
            Contract.Requires<ArgumentNullException>(targetType != null);
            try
            {
                if ( ! targetType.IsGenericType)
                {
                    return Convert.ChangeType(value, targetType);
                }
                var genericArgs = targetType.GetGenericArguments();
                if (genericArgs.Length != 1)
                {
                    throw new NotSupportedException
                        ("Generics with multiple arguments are not supported");
                }
                return Convert.ChangeType(value, genericArgs[0]);
            }
            catch (Exception)
            {
                return targetType.IsInstanceOfType(value) 
                    ? value 
                    : value.CloneConvert(targetType);
            }
        }

        public static T CloneConvert<T>(this object value)
        {
            return (T)CloneConvert(value, typeof (T));
        }

        public static object CloneConvert(this object value, Type targetType)
        {
            try
            {
                var result = Activator.CreateInstance(targetType);
                return CopyTo(value, result);
            }
            catch (Exception ex)
            {
                throw CreateFailedToConvertException(value, targetType, ex);
            }
        }

        /// <summary>
        /// Makes a shallow copy of source properties into target properties
        /// </summary>
        /// <returns>A object with a shallow copy of source properties</returns>
        public static object CopyTo(this object source, object target)
        {
            var targetType = target.GetType();
            var targetProps = targetType.GetProperties(BindingFlagsToUse);
            var targetFields = targetType.GetFields(BindingFlagsToUse);
            var sourceType = source.GetType();
            foreach (var targetPropertyInfo in targetProps.Where(propertyInfo => propertyInfo.CanWrite))
            {
                var sourceProperty = sourceType.GetProperty(targetPropertyInfo.Name, BindingFlagsToUse);
                if (sourceProperty == null 
                    ||  ! sourceProperty.CanRead)
                {
                    continue;
                }
                targetPropertyInfo.SetValue(target, sourceProperty.GetValue(source, null),
                                      null);
            }
            foreach (var fieldInfo in targetFields)
            {
                var sourceFieldInfo = sourceType.GetField(fieldInfo.Name, BindingFlagsToUse);
                if (sourceFieldInfo == null)
                {
                    continue;
                }
                fieldInfo.SetValue(target, sourceFieldInfo.GetValue(source));
            }
            return target;
        }

        /// <summary>
        /// Makes a shallow copy of source property values into null target properties
        /// </summary>
        public static object MergeTo(this object source, object target)
        {
            var targetType = target.GetType();
            var targetProps = targetType.GetProperties(BindingFlagsToUse);
            var targetFields = targetType.GetFields(BindingFlagsToUse);
            var sourceType = source.GetType();
            foreach (var targetPropertyInfo in targetProps
                .Where(propertyInfo => propertyInfo.CanWrite
                    && propertyInfo.CanRead
                    && propertyInfo.GetValue(target, null) == null))
            {
                var sourceProperty = sourceType.GetProperty(targetPropertyInfo.Name, BindingFlagsToUse);
                if (sourceProperty == null 
                    || ! sourceProperty.CanRead)
                {
                    continue;
                }
                var sourcePropertyValue = sourceProperty.GetValue(source, null);
                targetPropertyInfo.SetValue(target, sourcePropertyValue,
                                      null);
            }
            foreach (var fieldInfo in targetFields
                .Where(fieldInfo => fieldInfo.GetValue(target) == null))
            {
                var sourceFieldInfo = sourceType.GetField(fieldInfo.Name, BindingFlagsToUse);
                if (sourceFieldInfo == null)
                {
                    continue;
                }
                fieldInfo.SetValue(target, sourceFieldInfo.GetValue(source));
            }
            return target;
        }

        private static InvalidCastException CreateFailedToConvertException(object value, Type targetType, Exception originalException)
        {
            return new InvalidCastException(
                string.Format("Failed to convert type '{0}' into type '{1}'", value.GetType().FullName,
                              targetType.FullName), originalException);
        }
    }
}